import concurrent.futures
import enum
import re

import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    INT32_ELEMENT,
    KARABO_CLASSINFO,
    STRING_ELEMENT,
    TABLE_ELEMENT,
    ChannelMetaData,
    Hash,
    Schema,
    State,
)
from TrainMatcher import TrainMatcher

from . import shmem_utils, utils
from ._version import version as deviceVersion


class MergeGroupType(enum.Enum):
    MULTISOURCE = "sources"  # same key stacked from multiple sources in new source
    MULTIKEY = "keys"  # multiple keys within each matched source is stacked in new key


def merge_schema():
    schema = Schema()
    (
        BOOL_ELEMENT(schema)
        .key("select")
        .displayedName("Select")
        .assignmentOptional()
        .defaultValue(False)
        .reconfigurable()
        .commit(),

        STRING_ELEMENT(schema)
        .key("sourcePattern")
        .displayedName("Source pattern")
        .assignmentOptional()
        .defaultValue("")
        .reconfigurable()
        .commit(),

        STRING_ELEMENT(schema)
        .key("keyPattern")
        .displayedName("Key pattern")
        .assignmentOptional()
        .defaultValue("")
        .reconfigurable()
        .commit(),

        STRING_ELEMENT(schema)
        .key("replacement")
        .displayedName("Replacement")
        .assignmentOptional()
        .defaultValue("")
        .reconfigurable()
        .commit(),

        STRING_ELEMENT(schema)
        .key("type")
        .displayedName("Group type")
        .options(",".join(option.value for option in MergeGroupType))
        .assignmentOptional()
        .defaultValue(MergeGroupType.MULTISOURCE.value)
        .reconfigurable()
        .commit(),

        INT32_ELEMENT(schema)
        .key("stackingAxis")
        .displayedName("Axis")
        .assignmentOptional()
        .defaultValue(0)
        .reconfigurable()
        .commit(),
    )

    return schema


@KARABO_CLASSINFO("ShmemTrainMatcher", deviceVersion)
class ShmemTrainMatcher(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            TABLE_ELEMENT(expected)
            .key("merge")
            .displayedName("Array stacking")
            .allowedStates(State.PASSIVE)
            .description(
                "Specify which source(s) or key(s) to stack."
                "When stacking sources, the 'Source pattern' is interpreted as a "
                "regular expression and the 'Key pattern' is interpreted as an "
                "ordinary string. From all sources matching the source pattern, the "
                "data under this key (should be array with same dimensions across all "
                "stacked sources) is stacked in the same order as the sources are "
                "listed in 'Data sources' and the result is under the same key name in "
                "a new source named by 'Replacement'. "
                "When stacking keys, both the 'Source pattern' and the 'Key pattern' "
                "are regular expressions. Within each source matching the source "
                "pattern, all keys matching the key pattern are stacked and the result "
                "is put under the key named by 'Replacement'. "
                "While source stacking is optimized and can use thread pool, key "
                "stacking will iterate over all paths in matched sources and naively "
                "call np.stack for each key pattern. In either case, data that is used "
                "for stacking is removed from its original location (e.g. key is erased "
                "from hash)."
            )
            .setColumns(merge_schema())
            .assignmentOptional()
            .defaultValue([])
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected)
            .key("useThreadPool")
            .displayedName("Use thread pool")
            .allowedStates(State.PASSIVE)
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected)
            .key("enableKaraboOutput")
            .displayedName("Enable Karabo channel")
            .allowedStates(State.PASSIVE)
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),
        )

    def initialization(self):
        self._stacking_buffers = {}
        self._source_stacking_indices = {}
        self._source_stacking_sources = {}
        self._key_stacking_sources = {}
        self._prepare_merge_groups(self.get("merge"))
        super().initialization()
        self._shmem_handler = shmem_utils.ShmemCircularBufferReceiver()
        if self.get("useThreadPool"):
            self._thread_pool = concurrent.futures.ThreadPoolExecutor()
        else:
            self._thread_pool = None

        if not self.get("enableKaraboOutput"):
            # it is set already by super by default, so only need to turn off
            self.output = None

    def preReconfigure(self, conf):
        super().preReconfigure(conf)
        if conf.has("merge") or conf.has("sources"):
            self._prepare_merge_groups(conf["merge"])
        if conf.has("useThreadPool"):
            if self._thread_pool is not None:
                self._thread_pool.shutdown()
                self._thread_pool = None
            if conf["useThreadPool"]:
                self._thread_pool = concurrent.futures.ThreadPoolExecutor()
        if conf.has("enableKaraboOutput"):
            if conf["enableKaraboOutput"]:
                self.output = self._ss.getOutputChannel("output")
            else:
                self.output = None

    def _prepare_merge_groups(self, merge):
        source_group_patterns = []
        key_group_patterns = []
        # split by type, prepare regexes
        for row in merge:
            if not row["select"]:
                continue
            group_type = MergeGroupType(row["type"])
            if group_type is MergeGroupType.MULTISOURCE:
                source_group_patterns.append(
                    (
                        re.compile(row["sourcePattern"]),
                        row["keyPattern"],
                        row["replacement"],
                        row["stackingAxis"],
                    )
                )
            else:
                key_group_patterns.append(
                    (
                        re.compile(row["sourcePattern"]),
                        re.compile(row["keyPattern"]),
                        row["replacement"],
                    )
                )

        # not filtering by row["select"] to allow unselected sources to create gaps
        source_names = [row["source"].partition("@")[0] for row in self.get("sources")]

        self._stacking_buffers.clear()
        # handle source stacking groups
        self._source_stacking_indices.clear()
        self._source_stacking_sources.clear()
        for source_re, key, new_source, stack_axis in source_group_patterns:
            merge_sources = [
                source for source in source_names if source_re.match(source)
            ]
            if len(merge_sources) == 0:
                self.log.WARN(
                    f"Group merge pattern {source_re} did not match any known sources"
                )
                continue
            for (i, source) in enumerate(merge_sources):
                self._source_stacking_sources.setdefault(source, []).append(
                    (key, new_source, stack_axis)
                )
                self._source_stacking_indices[(source, new_source, key)] = i

        # handle key stacking groups
        self._key_stacking_sources.clear()
        for source_re, key_re, new_key in key_group_patterns:
            for source in source_names:
                # TODO: maybe also warn if no matches here?
                if not source_re.match(source):
                    continue
                self._key_stacking_sources.setdefault(source, []).append(
                    (key_re, new_key)
                )

    def _update_stacking_buffer(self, new_source, key, individual_shape, axis, dtype):
        # TODO: handle ValueError for max of empty sequence
        stack_num = (
            max(
                index
                for (
                    _,
                    new_source_,
                    key_,
                ), index in self._source_stacking_indices.items()
                if new_source_ == new_source and key_ == key
            )
            + 1
        )
        self._stacking_buffers[(new_source, key)] = np.empty(
            shape=utils.stacking_buffer_shape(individual_shape, stack_num, axis=axis),
            dtype=dtype,
        )

    def _handle_source(self, source, data_hash, timestamp, new_sources_map):
        # dereference calng shmem handles
        self._shmem_handler.dereference_shmem_handles(data_hash)

        # stack across sources (many sources, same key)
        # could probably save ~100 ns by "if ... in" instead of get
        for (stack_key, new_source, stack_axis) in self._source_stacking_sources.get(
            source, ()
        ):
            this_data = data_hash.get(stack_key)
            try:
                this_buffer = self._stacking_buffers[(new_source, stack_key)]
                stack_index = self._source_stacking_indices[
                    (source, new_source, stack_key)
                ]
                utils.set_on_axis(this_buffer, this_data, stack_index, stack_axis)
            except (ValueError, IndexError, KeyError):
                # ValueError: wrong shape (react to this_data.shape)
                # KeyError: buffer doesn't exist yet
                # IndexError: new source? (buffer not long enough)
                # either way, create appropriate buffer now
                # TODO: complain if shape varies between sources within train
                self._update_stacking_buffer(
                    new_source,
                    stack_key,
                    this_data.shape,
                    axis=stack_axis,
                    dtype=this_data.dtype,
                )
                # and then try again
                this_buffer = self._stacking_buffers[(new_source, stack_key)]
                stack_index = self._source_stacking_indices[
                    (source, new_source, stack_key)
                ]
                utils.set_on_axis(this_buffer, this_data, stack_index, stack_axis)
            # TODO: zero out unfilled buffer entries
            data_hash.erase(stack_key)

            if new_source not in new_sources_map:
                new_sources_map[new_source] = (Hash(), timestamp)
            new_source_hash = new_sources_map[new_source][0]
            if not new_source_hash.has(stack_key):
                new_source_hash[stack_key] = this_buffer

        # stack keys (multiple keys within this source)
        for (key_re, new_key) in self._key_stacking_sources.get(source, ()):
            # note: please no overlap between different key_re
            # note: if later key_re match earlier new_key, this gets spicy
            stack_keys = [key for key in data_hash.paths() if key_re.match(key)]
            try:
                # TODO: consider reusing buffers here, too
                stacked = np.stack([data_hash.get(key) for key in stack_keys], axis=0)
            except Exception as e:
                self.log.WARN(f"Failed to stack {key_re} for {source}: {e}")
            else:
                for key in stack_keys:
                    data_hash.erase(key)
                data_hash[new_key] = stacked

    def on_matched_data(self, train_id, sources):
        new_sources_map = {}
        if self._thread_pool is None:
            for source, (data, timestamp) in sources.items():
                self._handle_source(source, data, timestamp, new_sources_map)
        else:
            concurrent.futures.wait(
                [
                    self._thread_pool.submit(
                        self._handle_source, source, data, timestamp, new_sources_map
                    )
                    for source, (data, timestamp) in sources.items()
                ]
            )
        sources.update(new_sources_map)

        # karabo output
        if self.output is not None:
            for source, (data, timestamp) in sources.items():
                self.output.write(data, ChannelMetaData(source, timestamp))
            self.output.update()

        # karabo bridge output
        for source, (data, timestamp) in sources.items():
            self.zmq_output.write(source, data, timestamp)
        self.zmq_output.update()

        self.info["sent"] += 1
        self.info["trainId"] = train_id
        self.rate_out.update()
