import enum
import collections
import functools
import inspect
import threading
import time
from timeit import default_timer

import numpy as np


def pick_frame_index(selection_mode, index, cell_table, pulse_table, warn_func=None):
    """When selecting a single frame to preview, an obvious question is whether the
    number the operator provides is a frame index, a cell ID, or a pulse ID. This
    function allows any of the three, translating into frame index.

    As this will be used by correction devices, the warn_func parameter allows the
    function to issue warnings via this instead of raising exceptions.

    Indices below zero are special values and thus returned directly.

    Returns: (found frame index, corresponding cell ID, corresponding pulse ID)"""

    if index < 0:
        return index, index, index

    # TODO: enum
    if selection_mode == "frame":
        if index >= cell_table.size:
            if warn_func is not None:
                warn_func(
                    f"Index {index} out of range for cell table of length "
                    f"{len(cell_table)}, returning index 0 instead"
                )
            frame_index = 0
        else:
            frame_index = index

        return frame_index, cell_table[frame_index], pulse_table[frame_index]
    elif selection_mode == "cell":
        found = np.where(cell_table == index)[0]
        if len(found) > 0:
            cell = index
            frame_index = found[0]
        else:
            cell = cell_table[0]
            if warn_func is not None:
                warn_func(
                    f"Cell {index} not found, arbitrary cell {cell} returned instead"
                )
            frame_index = 0
        return frame_index, cell, pulse_table[frame_index]
    elif selection_mode == "pulse":
        found = np.where(pulse_table == index)[0]
        if len(found) > 0:
            pulse = index
            frame_index = found[0]
        else:
            pulse = pulse_table[0]
            if warn_func is not None:
                warn_func(
                    f"Pulse {index} not found, arbitrary pulse {pulse} returned instead"
                )
            frame_index = 0
        return frame_index, cell_table[frame_index], pulse
    else:
        raise ValueError(f"Invalid selection mode '{selection_mode}'")


def threadsafe_cache(fun):
    """This decorator imitates functools.cache, but threadsafer

    With multiple threads hitting a function cached by functools.cache, it is possible
    to trigger recomputation. This decorator adds granular locking: each key in the
    cache (derived from arguments) has its own lock.
    """

    locks = {}
    results = {}
    fun_sig = inspect.signature(fun)

    @functools.wraps(fun)
    def aux(*args, **kwargs):
        bound_args = fun_sig.bind(*args, **kwargs)
        bound_args.apply_defaults()
        key = bound_args.args + tuple(bound_args.kwargs.items())
        if key in results:
            return results[key]
        with locks.setdefault(key, threading.Lock()):
            if key in results:
                # someone else did this - may still be processing
                return results[key]
            else:
                res = fun(*args, **kwargs)
                results[key] = res
                return res

    return aux


@functools.lru_cache()
def transpose_order(axes_in, axes_out):
    """Computes the order of axes_out relative to axes_in for transposition purposes

    Both axes_in and axes_out are assumed to be strings in which each letter represents
    an axis (duck typing accepts: any iterable of hashable elements). They should
    probably be of the same length and have no repetitions, but this is not enforced.
    Off-label use voids warranty.
    """
    axis_order = {axis: index for index, axis in enumerate(axes_in)}
    return tuple(axis_order[axis] for axis in axes_out)


def stacking_buffer_shape(array_shape, stack_num, axis=0):
    """Figures out the shape you would need for np.stack"""
    if axis > len(array_shape) or axis < -len(array_shape) - 1:
        # complain when np.stack would
        raise np.AxisError(
            f"axis {axis} is out of bounds "
            f"for array of dimension {len(array_shape) + 1}"
        )
    if axis < 0:
        axis += len(array_shape) + 1
    return array_shape[:axis] + (stack_num,) + array_shape[axis:]


def set_on_axis(array, vals, index, axis):
    """set_on_axis(A, x, 1, 2) corresponds to A[:, :, 1] = x"""
    if axis >= len(array):
        raise IndexError(
            f"too many indices for array: array is {len(array.shape)}-dimensional, "
            f"but {axis+1} were indexed"
        )
    # TODO: maybe support negative axis with wraparound
    indices = np.index_exp[:] * axis + np.index_exp[index]
    array[indices] = vals


_np_typechar_to_c_typestring = {
    "?": "bool",
    "B": "unsigned char",
    "D": "double complex",
    "F": "float complex",
    "G": "long double complex",
    "H": "unsigned short",
    "I": "unsigned int",
    "L": "unsigned long",
    "Q": "unsigned long long",
    "b": "char",
    "d": "double",
    "e": "half",  # warning: only in CUDA with special support
    "f": "float",
    "g": "long double",
    "h": "short",
    "i": "int",
    "l": "long",
    "q": "long long",
}


def np_dtype_to_c_type(dtype):
    as_char = np.sctype2char(dtype)
    return _np_typechar_to_c_typestring[as_char]


def enum_to_c_template(enum_class):
    res = [f"enum {enum_class.__name__} {{"]
    for field in enum_class:
        res.append(f"\t{field.name} = {field.value},")
    res.append("};")
    return "\n".join(res)


def ceil_div(num, denom):
    return (num + denom - 1) // denom


class RepeatingTimer:
    """A timer which will call callback every interval seconds"""

    def __init__(
        self,
        interval,
        callback,
        start_now=True,
        daemon=True,
    ):
        self.stopped = True
        self.interval = interval
        self.callback = callback
        self.daemonize = daemon
        if start_now:
            self.start()

    def start(self):
        self.stopped = False
        self.wakeup_time = default_timer() + self.interval

        def runner():
            while not self.stopped:
                now = default_timer()
                while now < self.wakeup_time:
                    diff = self.wakeup_time - now
                    time.sleep(diff)
                    if self.stopped:
                        return
                    now = default_timer()
                self.callback()
                self.wakeup_time = default_timer() + self.interval

        self.thread = threading.Thread(target=runner, daemon=self.daemonize)
        self.thread.start()

    def stop(self):
        self.stopped = True


class ExponentialMovingAverage:
    def __init__(self, alpha, use_first_value=True):
        self.alpha = alpha
        self.initialised = not use_first_value
        self.mean = 0

    def update(self, value):
        if self.initialised:
            self.mean += self.alpha * (value - self.mean)
        else:
            self.mean = value
            self.initialised = True

    def get(self):
        return self.mean


class WindowRateTracker:
    def __init__(self, buffer_size=20, time_window=10):
        self.time_window = time_window
        self.buffer_size = buffer_size
        self.deque = collections.deque(maxlen=self.buffer_size)

    def update(self):
        self.deque.append(default_timer())

    def get(self):
        now = default_timer()
        cutoff = now - self.time_window
        try:
            while self.deque[0] < cutoff:
                self.deque.popleft()
        except IndexError:
            return 0
        if len(self.deque) < 2:
            return 0
        if len(self.deque) < self.buffer_size:
            # TODO: estimator avoiding ramp-up of when starting anew
            return len(self.deque) / self.time_window
        else:
            # if going faster than buffer size per time window, look at timestamps
            oldest, newest = self.deque[0], self.deque[-1]
            buffer_span = newest - oldest
            period = buffer_span / (self.buffer_size - 1)
            if (now - newest) < period:
                # no new estimate yet, expecting new event after period
                return 1 / period
            else:
                return self.buffer_size / (now - oldest)


class Stopwatch:
    """Context manager measuring time spent in context.

    Keyword arguments:
    name: if not None, will appear in string representation
          also, if not None, will automatically print self when done
    """

    def __init__(self, name=None):
        self.stop_time = None
        self.name = name

    def __enter__(self):
        self.start_time = default_timer()
        return self

    def __exit__(self, t, v, tb):  # type, value and traceback irrelevant
        self.stop_time = default_timer()
        if self.name is not None:
            print(repr(self))

    @property
    def elapsed(self):
        if self.stop_time is not None:
            return self.stop_time - self.start_time
        else:
            return default_timer() - self.start_time

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        if self.name is None:
            return f"{self.elapsed():.3f} s"
        else:
            return f"{self.name}: {self.elapsed():.3f} s"


class StateContext:
    """What if device state was a stack?"""

    def __init__(self, device, new_state, revert_to=None):
        self.device = device
        if revert_to is None:
            revert_to = device.get("state")
        self.revert_to = revert_to
        self.new_state = new_state

    def __enter__(self):
        if self.new_state is not self.revert_to:
            self.device.updateState(self.new_state)

    def __exit__(self, t, v, tb):
        if self.new_state is not self.revert_to:
            self.device.updateState(self.revert_to)


class TrainRatioTracker:
    """Measure how many percent of recent train IDs (from contiguous set) were seen

    The tracker will maintain a queue of buffer_size train IDs going back at most
    buffer_size from latest train ID (depending on calls to get). Call update(train_id)
    when you see a new train and call get to get() the ratio of recent trains seen.

    In case warn_callback is given, update can issue a warning in case invalid train
    IDs are received. The tracker assumes trains are strictly increasing and that they
    are supposed to be contiguous - hence the ability to infer when some are missing.
    """

    def __init__(self, buffer_size=50, warn_callback=None):
        self._train_id_queue = collections.deque(maxlen=buffer_size)
        self._train_id_range = buffer_size
        self._warn_callback = warn_callback

    def get(self, current_train=None):
        """Get the ratio of recent trains until current_train or latest updated.

        If current_train is not provided, then the range considered in computing the
        ratio will be from latest train_id in update (going back buffer_size trains).

        If you happen to know a more current train ID that has not been given in a
        call to update (maybe you are receiving invalid trains and don't count them),
        you can give this as current_train, yielding a lower ratio. Note that this will
        trim the queue, so a subsequent call with lower or no current_train will return
        an incorrectly low ratio.
        """
        if len(self._train_id_queue) == 0:
            return 0

        if current_train is None:
            current_train = self._train_id_queue[-1]

        cutoff = current_train - self._train_id_range + 1
        try:
            while self._train_id_queue[0] < cutoff:
                self._train_id_queue.popleft()
        except IndexError:
            return 0

        # TODO: avoid estimator ramp-up (don't initially divide by full range)
        return len(self._train_id_queue) * 100 / self._train_id_range

    def update(self, train_id):
        if (
            len(self._train_id_queue) > 0
            and self._train_id_queue[-1] >= train_id
            and self._warn_callback is not None
        ):
            self._warn_callback(
                f"New train ID {train_id} not greater than last thing in queue, "
                f"{self._train_id_queue[-1]}, just thought you should know..."
            )
        self._train_id_queue.append(train_id)


class ChainHash:
    """Like read-only ChainMap, but for karabo.bound.Hash(es) instead!"""

    def __init__(self, *hashes):
        self._hashes = hashes

    def __getitem__(self, key):
        for h in self._hashes:
            if h.has(key):
                return h[key]
        raise KeyError()


class SkippingThrottler:
    def __init__(self, min_period):
        self.min_period = min_period
        self.latest_ts = float("-inf")
        self.lock = threading.Lock()

    def test_and_set(self):
        with self.lock:
            now = default_timer()
            if (now - self.latest_ts) >= self.min_period:
                self.latest_ts = now
                return True
            else:
                return False


class BadPixelValues(enum.IntFlag):
    """The European XFEL Bad Pixel Encoding

    Straight from pycalibration's enum.py"""

    OFFSET_OUT_OF_THRESHOLD = 2 ** 0
    NOISE_OUT_OF_THRESHOLD = 2 ** 1
    OFFSET_NOISE_EVAL_ERROR = 2 ** 2
    NO_DARK_DATA = 2 ** 3
    CI_GAIN_OUT_OF_THRESHOLD = 2 ** 4
    CI_LINEAR_DEVIATION = 2 ** 5
    CI_EVAL_ERROR = 2 ** 6
    FF_GAIN_EVAL_ERROR = 2 ** 7
    FF_GAIN_DEVIATION = 2 ** 8
    FF_NO_ENTRIES = 2 ** 9
    CI2_EVAL_ERROR = 2 ** 10
    VALUE_IS_NAN = 2 ** 11
    VALUE_OUT_OF_RANGE = 2 ** 12
    GAIN_THRESHOLDING_ERROR = 2 ** 13
    DATA_STD_IS_ZERO = 2 ** 14
    ASIC_STD_BELOW_NOISE = 2 ** 15
    INTERPOLATED = 2 ** 16
    NOISY_ADC = 2 ** 17
    OVERSCAN = 2 ** 18
    NON_SENSITIVE = 2 ** 19
    NON_LIN_RESPONSE_REGION = 2 ** 20
